#!/usr/bin/perl

# $Id: spamcheck.milter.pl,v 1.16 2003/04/20 04:23:35 dredd Exp $
#
# There's a bug in this thing SOMEWHERE that causes it to lock up tighter
# than a drum when using tagging. I'm not sure what the deal is, but every 
# so often it just wedges. If you find the bug, you get the credit
#
# This also now uses the {client_resolve} macro, which means you need to 
# configure your sendmail.mc to also include it on the connect callback
# typically via something like
#
#    define(`confMILTER_MACROS_CONNECT',``j, _, {daemon_name}, {if_name}, \ 
#        {if_addr}, {client_resolve}'')dnl
#
# (that's a single line, in case the "\" doesn't tip you off)
#
# Release History
#
# 1.0b3  - 20 April 2003
#    o  Fixed logging bug where mx_forbidden was showing up in logs
#       as freemail ... copy/paste sucks.
# 1.0b2  - 19 April 2003
#    o  Added mx_forbidden check, to check for RFC1918 or other broken MX
#       entries.
#    o  removed use of $DEBUG, converted to using a constant DEBUG like
#       should have been done in the first place.
#
# 1.0b1  - 30 March 2003
#

use English;
use Net::CIDR;
use Net::DNS;
use Sendmail::AccessDB qw(spam_friend whitelisted lookup);
use Sendmail::Milter;
use Socket;
use Sys::Syslog qw(:DEFAULT setlogsock);
use vars qw($regex_lock);
use re 'eval';
use DB_File;
use strict;
use constant DEBUG=>1;

$0 = 'Sendmail Milter - SpamCheck';
$OUTPUT_AUTOFLUSH = 1;

my $FACILITY = 'local4';

my $PID_FILE = '/var/run/spamcheck.pid';

#      "TESTING" or "AS YET UNIMPLEMENTED" features

# 'mx_forbidden' (TESTING)
#    If an MX is either in a broken format (e.g., an IP address instead
#    of a hostname), or sitting in RFC1918 space

#      "STABLE" FEATURES

# 'helo_not_resolve'
#    will reject mail if gethostbyname() doesn't yield at least one entry
# 'known_bad_helo'
#    insert common spamsign helo arguments (YOUR hostname, for example)
#    to reject mail with that spamsign.
# 'access_db'
#    move access_db checks into the milter so they can be logged better
# 'connect_ip_dnsbl'
# 'connect_hostname_dnsbl'
#    blacklists based on IP or domain, respectively, of the connecting
#    host (replacement for dnsbl.m4) [not yet implemented]
# 'sender_rhs'
#    do lookups against the RHS of the MAIL sender argument. not needed if
#    using rhsbl.m4 
# 'helo_argument'
#    do lookups against the helo argument - Caution, doing this directly 
#    violates RFC2821/4.1.4 if used for the purpose of message rejection 
#    (tagging would be fine)
# 'spam_urls'
#    Do lookups on the host-parts of URL's. 
# 'from_headers'
#    Deals with the "From:" header, not the envelope, checking the RHS of the
#    address
# 'reply_to_headers'
#    Deals with the "Reply-To:" header, not the envelope, checking the RHS
#    of the address
# 'unresolvable_urls'
#    Checks to see if a numeric IP actually resolves to a name (common spam
#    technique is to serve up on rDNS-lacking host)
# 'obfuscated_urls'
#    Checks for intentionally obfuscated, but valid, host parts.
# 'freemail' 
#    Checks against known freemail providers to ensure they're coming from
#    where they should be.
# 'known_bad_subjects'
#    Scan subject for known bad regexes (e.g. "ADV:")
# 'host_bad_ns'
# 'sender_bad_ns'
# 'from_bad_ns'
# 'reply_to_bad_ns'
# 'url_bad_ns'
#    consult a list of "known bad" NS servers to see if a domain contained
#    in the hostname/sender/headers/URLs is associated with the domains of a 
#    "known spamhaus"
# 'bad_message_id'
#    Compare the message ID header to the regex at the top of this program
# 'missing_message_id'
#    Reject/tag messages with no message ID
# 'body_bad_regex'
#    Look for common "spam signs" in the body of the message itself.
# 'client_resolve_forged'
# 'client_resolve_temp'
# 'client_resolve_fail'
#    Handle the 3 possible failure modes for DNS failure. 'client_resolve_temp'
#    if set to 'reject', will in fact reject with a 450 temporary condition, 
#    since the error condition causing the problem is temporary.
# 'bad_headers'
#    List of headers that, if present, are cause for rejection (e.g., an
#    X-UIDL header should not appear in a message).


my $SMTP_AUTH_CAN_DO_NO_WRONG = 1;     # If set, folks who have authenticated
                                       # via SASL will be automatically
                                       # accepted. Use this only if you're
                                       # _not_ trying to filter your users
                                       # outbound mail.

my $CLIENT_RESOLV_ALWAYS_TMPFAIL = 0;  # Set this, any any "rejects" for 
                                       # client_resolv will always use 451
                                       # response codes. If unset
                                       # then only tmpfail failures will
                                       # tmpfail


my $SPAMLOG_ALL_TMPFAILS = 0;   # if set, and if CLIENT_RESOLV_ALWAYS_TMPFAIL
				# is also set, log messages for client_resolv
				# failures (no reverse DNS and possible 
				# forged DNS) are logged to syslog along
				# real rejections, even though they return
				# SMFIS_TEMPFAIL.

my $ALLOW_THIRD_PARTY_SENDERS = 1; # Turning this option on will cause
                                   # the elements in the array
                                   # @freemail_exceptions to be used as
                                   # being globally accepted sources of
                                   # freemail domains (e.g., paypal.com,
                                   # ebay.com, americangreetings.com, etc.)

# Local info about users with forwards

my %users_with_forwards = (
			   'bc@megacity.org' => ['netnation.com'], 
                           'bc@bjb.org' => ['netnation.com']
                           );

my @backup_mx_hosts = ();  # Needed for freemail, if you have any backup
                           # MX's that might ALSO be delivering their mail

# Default information

my $SPAMCHECK_FILES = "/etc/mail/spamcheck";

# Freemail information

my %freemail_relays = ( 
                        'yahoo.com' => [ '.yahoo.com', '.yahoomail.com' ],
                        'aol.com' =>   [ '.aol.com' ],
                        'juno.com' =>  [ '.juno.com' ],
                        'hotbot.com' =>[ '.hotbot.com', '.mailcity.com' ],
                        'msn.com' =>   [ '.msn.com', '.hotmail.com', 
                                         'microsoft.com'],
                        'hotmail.com'=>[ '.msn.com', '.hotmail.com', 
                                         '.microsoft.com']
                        );

#
# We will accept freemail relayed from these hosts, since it actually
# originates there, but gives the users email as the return address.
# (Kinda like a third party introduction system.)

my @freemail_exceptions;
if ($ALLOW_THIRD_PARTY_SENDERS)
{
    @freemail_exceptions = (
			    'ebay.com',
			    'paypal.com'
			    );
}

# Forbidden exceptions is a list of domains (RHS) that will not be checked
# for MX-in-reserved-space violations (use this for your own domain, for
# example, if you have DNS reflecting private address space, or for
# partners which may have shared RFC1918 space with you

my @forbidden_exceptions = ();

#
#
#
# Beyond this point you should not have to edit.
#
#
#

# Define the reserved space

my @forbidden = ('0.0.0.0/8','127.0.0.0/8','10.0.0.0/8','172.16.0.0/12',
		 '169.254.0.0/16','192.0.2.0/24','192.88.99.0/24',
		 '192.168.0.0/16','198.18.0.0/15','224.0.0.0/4');

# Jeff Friedl's regex stuff. He rules.

my $WSP = '[ \t]';

# FWS     =       ([*WSP CRLF] 1*WSP) / obs-FWS
# obs-FWS =       1*WSP *(CRLF 1*WSP)
my $FWS = qr{(?>
                (?: $WSP* \015\012 )? $WSP+   # FWS
                |
                $WSP+ (?: \015\012 $WSP+ )*   # obs-FWS
            )}x;

our $OpenParens;
my $NestedParenGuts = qr{

    (?{ local $OpenParens = 0 }) ## counts the number of nested opens waiting to close

    (?>
     (?>
        ## stuff not parens, not escaped
         [^()\\]+

        ## escaped stuff
        | (?s: \\. )

        # another opening paren
        | \( (?{ $OpenParens++ })

        # a closing paren, if we're expecting any
			   | (?(?{ $OpenParens }) (?{ $OpenParens-- }) \))
      )*
     )
}x;


my $COMMENT = qr{(?>
                 \(
                   $NestedParenGuts
                 \)
    )}x;

# CFWS =       *([FWS] comment) (([FWS] comment) / FWS)
# which seems to me to be the same as
# CFWS =       +([FWS] comment) [FWS] / FWS
my $CFWS = qr{(?>
                (?: $FWS? $COMMENT )+ $FWS?
                |
                $FWS?
    )}x;

my $valid_message_id_regex = qr{
   ^
   $CFWS?
   <
     (?:
          "(?>[^\\\"]|\\.)*"
          |
          (?> [-a-zA-Z0-9!\#\$%&\'*+/=?^_`{|}~]+  \. )*
              [-a-zA-Z0-9!\#\$%&\'*+/=?^_`{|}~]+
     )
     \@
     (?>
        \[ (?>[^\\\]]|\\.)*  \]
        |
        (?> [-a-zA-Z0-9!\#\$%&\'*+/=?^_`{|}~]+ \. )*
            [-a-zA-Z0-9!\#\$%&\'*+/=?^_`{|}~]+
     )
   >
   $CFWS?
   $
  }x;



sub connect_callback
{
    my $ctx = shift;
    my $hostname = shift;
    my $sockaddr_in = shift;
    my ($port, $iaddr);

    print "connect_callback:\n" if DEBUG;
    print "   + hostname: '$hostname'\n" if DEBUG;
    my $client_resolve = $ctx->getsymval('{client_resolve}');
    print "client_resolve: $client_resolve\n" if DEBUG;

    ($port, $iaddr) = sockaddr_in($sockaddr_in);
    my $ip_addr = inet_ntoa($iaddr);

    print "IP Address is : $ip_addr\n" if DEBUG;

    if ( 
	 (whitelisted($ip_addr))
	 or
	 (whitelisted($ip_addr, 'qualifier'=>'Connect'))
	 )
    {	
	print "IP address $ip_addr explicitly (globally) whitelisted.\n" if DEBUG;
	return SMFIS_ACCEPT;
    }

    print "Got here.\n" if DEBUG;
    
    if ( (whitelisted($hostname)) or
	 (whitelisted($hostname,'qualifier'=>'Connect')) )
    {
	print "Host $hostname explicitly (globally) whitelisted.\n" if DEBUG;
	return SMFIS_ACCEPT;
    }

    my %hash = ( 'host' => $hostname,
		 'client_resolve' => $client_resolve,
		 'ip_addr' => $ip_addr);
    print "Setting initial private data cache, containing hostname\n" if DEBUG;

    $ctx->setpriv(\%hash);

    print "+ CONNECT RETURNING\n" if DEBUG;
    return SMFIS_CONTINUE;
}

sub helo_callback
{
    my $ctx = shift;
    my $helo_hostname = shift;

    print "helo_callback:\n" if DEBUG;

    if ( ($helo_hostname !~ /^\[(?:\d{1,3}\.){3}\d{1,3}\]$/)
	 and
	 ($helo_hostname !~ 
	  /^(?:[A-Za-z0-9]+(?:-+[A-Za-z0-9]+)*(?:\.[A-Za-z0-9]+(?:-+[A-Za-z0-9]+)*)+)$/
	  )
	 )
    {
	print "Has a bad HELO argument '$helo_hostname'\n" if DEBUG;
	my $reason = "Message rejected because the HELO argument ($helo_hostname) is not a valid FQDN or dotted-quad. You probably need to fix your mail server's broken configuration.";
	my $href = $ctx->getpriv();
	$$href{'helo_argument'} = $reason;
	$ctx->setpriv($href);
    }

    my @known_bad_helos = user_config('','known_bad_helos');

	
    if (grep /^$helo_hostname$/i, @known_bad_helos)
    {
	print "Has a known-bad HELO argument '$helo_hostname'\n" if DEBUG;
	my $reason = "Message rejected because the HELO argument ($helo_hostname) is demonstrably false or a likely spamsign.";
	my $href = $ctx->getpriv();
	$$href{'known_bad_helo'} = $reason;
	$ctx->setpriv($href);
    }

    if ($helo_hostname =~ /^(?:[A-Za-z0-9]+(?:-+[A-Za-z0-9]+)*(?:\.[A-Za-z0-9]+(?:-+[A-Za-z0-9]+)*)+)$/)
    {
	# this is potentially a valid hostname
	if (! gethostbyname($helo_hostname))
	{
	    print "HELO argument ($helo_hostname) does not resolve.\n" if DEBUG;
	    my $reason = "Message rejected because the HELO argument ($helo_hostname) does not resolve. Your mailserver probably needs to be reconfigured to fix that.";
	    my $href = $ctx->getpriv();
	    $$href{'helo_not_resolve'} = $reason;
	    $ctx->setpriv($href);
	}
    }
    
    print "+ RETURNING FROM HELO\n" if DEBUG;
    return SMFIS_CONTINUE;
}

sub from_callback
{
    my $ctx = shift;
    my @args = @_;

    print "+ from_callback\n" if DEBUG;

    my $auth_authen = $ctx->getsymval('{auth_authen}');
    if ( (defined $auth_authen) and
	 ( $SMTP_AUTH_CAN_DO_NO_WRONG ) 
	 )
    {
	print "SMTP/AUTH: $auth_authen is an authenticated sender. Anything goes.\n" if DEBUG;
	return SMFIS_ACCEPT;
    }

    my $sender = $args[0];

    my $clean_sender = lc $sender;
    $clean_sender =~ s/^\s*\<//;
    $clean_sender =~ s/\>\s*$//;
    
    if (whitelisted($clean_sender))
    {
	print "$clean_sender is explicitly (globally) whitelisted.\n" if DEBUG;
	return SMFIS_ACCEPT;
    }

    print "$clean_sender is normal... no special treatment.\n" if DEBUG;

    my $href = $ctx->getpriv();
    $$href{'clean_sender'} = $clean_sender;
    $$href{'dirty_sender'} = $sender;
    delete $$href{'already_got_a_rcpt'};
    $ctx->setpriv($href);
    
    my $rhs = $clean_sender;
    $rhs =~ s/.*\@//;

    ### Check sender RHS against freemail provider maps.

    my $effective_rhs = $rhs;

    my $connecting_host = $$href{'host'};
    
    if (
	($rhs =~ /^yahoo\.[A-Za-z]{2}$/i) or
	($rhs =~ /^yahoo\.com?\.[A-Za-z]{2}$/i) or
	($rhs =~ /^yahoo\-inc\.com$/)
	)
    {
	$effective_rhs = 'yahoo.com';
    }
    
    if (defined $freemail_relays{$effective_rhs})
    {
	print "$effective_rhs is a freemail provider.\n" if DEBUG;
	my $acceptable = 0;
	foreach my $host ( @{$freemail_relays{$effective_rhs}}, 
			   @backup_mx_hosts, @freemail_exceptions )
	{
	    print "Comparing /$connecting_host/ to /$host/\n" if DEBUG;
	    if ($connecting_host =~ /$host$/i)
	    {
		$acceptable = 1;
	    }
	}
	
	if (! $acceptable)
	{
	    print "We'd reject messages like this, but we'll store it for later use in case the recip likes spam.\n" if DEBUG;
	    # Rejection Code
	    my $reason = "Message rejected because mail from $clean_sender coming from $connecting_host does not actually appear to be from $rhs";
	    $$href{'freemail'} = $reason;
	    $ctx->setpriv($href);
	}
	
    }

    if (! grep /^$rhs$/i, @forbidden_exceptions)
    {
	print "Checking $rhs for a bad MX\n" if DEBUG;
	my ($rc,$rc_string) = bad_mx_set($rhs);
	if ($rc)
	{
	    print "$rhs has a bad MX ($rc_string). Making a note of that for later.\n" if DEBUG;
	    $$href{'mx_forbidden'} = [$rc, $rc_string];
	    $ctx->setpriv($href);
	}
    }

    print "+ RETURNING FROM FROM\n" if DEBUG;
    return SMFIS_CONTINUE;
}

sub rcpt_callback
{
    my $ctx = shift;
    my @args = @_;
    
    if (DEBUG)
    {
	print "rcpt_callback:\n";
	print "   + args: '" . join(', ', @args) . "'\n";
    }

    my $recip = lc $args[0];
    $recip =~ s/^\<//;
    $recip =~ s/\>$//;

    my $end_user = $ctx->getsymval('{rcpt_addr}');
    print "end_user is $end_user\n" if DEBUG;

    $end_user = alias_scan($end_user);
    print "This belongs to local user '$end_user'\n" if DEBUG;

    my $href = $ctx->getpriv();
    $ctx->setpriv($href);

    my $clean_sender = $$href{'clean_sender'};
    my $dirty_sender = $$href{'dirty_sender'};
    my $hostname = $$href{'host'};
    my $ip_addr = $$href{'ip_addr'};

    if (defined $$href{'already_got_a_rcpt'})
    {
	print "We already have received a RCPT for this message.\n" if DEBUG;
	$ctx->setreply('451','4.7.1','Only one RCPT per connection so that user-specific body filtering can be done. Please resend.');
	return SMFIS_TEMPFAIL;
    }
    else
    {
	print "$recip is the first RCPT for this message.\n" if DEBUG;
	$$href{'already_got_a_rcpt'} = $recip;
	$$href{'end_user'} = $end_user;
	$ctx->setpriv($href);
    }

    print "Checking $recip to see if he's a spam-friend...\n" if DEBUG;
    my $user_pref = spam_friend($recip);

    if ($user_pref eq 'FRIEND')
    {
	print "Accepted mail because $recip is a spam-friend.\n" if DEBUG;
	return SMFIS_ACCEPT;
    }
    else
    {
	print "nope. just a normal guy.\n" if DEBUG;
    }

    # whitelist checking

    print "Checking <$clean_sender> in access_db for $end_user\n" if DEBUG;
    my $wl_file = access_file($end_user,'whitelist');
    my $bl_file = access_file($end_user,'block');

    print "...using $wl_file\n" if DEBUG;

    my $sender_wl_access = lookup($clean_sender, 
				  'type' => 'mail',
				  'file' => $wl_file,
				  'qualifier' => 'from'
				  );

    print "Checking without 'from' qualifier'\n" if DEBUG;

    $sender_wl_access = lookup($clean_sender, 
			       'type' => 'mail',
			       'file' => $wl_file
			       )
	if ! defined $sender_wl_access;

    return SMFIS_ACCEPT if $sender_wl_access eq 'OK';
    
    print "Checking $ip_addr in access_db\n" if DEBUG;

    my $ip_wl_access = lookup($ip_addr,
			      'type' => 'ip',
			      'file' => $wl_file,
			      'qualifier' => 'connect'
			      );

    $ip_wl_access = lookup($ip_addr,
			   'type' => 'ip',
			   'file' => $wl_file
			   )
	if ! defined $ip_wl_access;
    
    return SMFIS_ACCEPT if $ip_wl_access eq 'OK';
        
    if (defined $hostname)
    {
	print "Checking $hostname in access_db\n" if DEBUG;

	my $host_wl_access = lookup($hostname,
				    'type' => 'hostname',
				    'file' => $wl_file,
				    'qualifier' => 'connect'
				    );
	
	$host_wl_access = lookup($hostname,
				 'type' => 'hostname',
				 'file' => $wl_file
				 )
	    if ! defined $host_wl_access;

	return SMFIS_ACCEPT if $host_wl_access eq 'OK';
    }

    my %rcpt_config = user_config($end_user,'*');

    print "Got past whitelist checks without matching.\n" if DEBUG;

    # Sender Syntax Error

    if ( (behavior('sender_syntax_error',$end_user) ne 'ignore') and
	 (is_valid_path($dirty_sender) == 0) )
    {
	print "$dirty_sender is not syntactically correct.\n" if DEBUG;
	if (behavior('sender_syntax_error',$end_user) eq 'reject')
	{
	    my $reason = "Envelope sender '$dirty_sender' is invalid syntax";
	    $ctx->setreply(550,'5.7.1',$reason);
	    syslog('notice','%s/%s 550 %s <%s> sender_syntax_error "%s"', 
		   $ip_addr, $hostname, $dirty_sender, $recip, 
		   $reason);
	    return SMFIS_REJECT;
	}
	else
	{
	    my $tag_name = 'sender_syntax_error';
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	}
    }

    print "Checking sender against block lists...\n" if DEBUG;

    # Access.db checks

    if (behavior('access_db',$end_user) ne 'ignore')
    {
	print "Checking <$clean_sender> in access_db\n" if DEBUG;
	my $sender_access = lookup($clean_sender, 
				   'type' => 'mail',
				   'file' => $bl_file,
				   'qualifier' => 'from'
				   );

	$sender_access = lookup($clean_sender, 
				'type' => 'mail',
				'file' => $bl_file
				)
	    if ! defined $sender_access;

	if (defined $sender_access)
	{
	    print "$clean_sender has status of $sender_access\n" if DEBUG;

	    if ($sender_access =~ /^ERROR/i)
	    {
		if (behavior('access_db',$end_user) eq 'reject')
		{
		    $sender_access =~ s/.*\"\d+\s*//;
		    $sender_access =~ s/\".*//;
		    $ctx->setreply(550,'5.7.1',$sender_access);
		    syslog('notice','%s/%s 550 <%s> <%s> access_db "%s"', 
			   $ip_addr, $hostname, $clean_sender, $recip, 
			   $sender_access);
		    return SMFIS_REJECT;
		}
		elsif (behavior('access_db',$end_user) eq 'tag')
		{
		    my $tag_name = 'access_db';
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}

	print "Checking <$ip_addr> in access_db\n" if DEBUG;
				
	my $ip_access = lookup($ip_addr,
			       'type' => 'ip',
			       'file' => $bl_file,
			       'qualifier' => 'connect'
			       );
	$ip_access = lookup($ip_addr,
			    'type' => 'ip',
			    'file' => $bl_file
			    )
	    if ! defined $ip_access;

	if (defined $ip_access)
	{
	    print "$ip_addr has status of $ip_access\n" if DEBUG;
	    if ($ip_access =~ /^ERROR/i)
	    {
		print "We'll reject this.\n" if DEBUG;
		if (behavior('access_db',$end_user) eq 'reject')
		{
		    $ip_access =~ s/.*\"\d+\s*//;
		    $ip_access =~ s/\".*//;
		    $ctx->setreply(550,'5.7.1',$ip_access);
		    syslog('notice','%s/%s 550 <%s> <%s> access_db "%s"',
			   $ip_addr, $hostname, $clean_sender, $recip,
			   $ip_access);
		    return SMFIS_REJECT;
		}
		elsif (behavior('access_db',$end_user) eq 'tag')
		{
		    my $tag_name = 'access_db';
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}

	print "Checking <$hostname> in access_db\n" if DEBUG;

	if (defined $hostname)
	{
	    my $host_access = lookup($hostname,
				     'type' => 'hostname',
				     'file' => $bl_file,
				     'qualifier' => 'connect'
				     );
	    
	    $host_access = lookup($hostname,
				  'type' => 'hostname',
				  'file' => $bl_file
				  )
		if ! defined $host_access;
	    
	    if (defined $host_access)
	    {
		print "$hostname has status of $host_access\n" if DEBUG;
		
		if ($host_access =~ /^ERROR/i)
		{
		    if (behavior('access_db',$end_user) eq 'reject')
		    {
			$host_access =~ s/.*\"\d+\s*//;
			$host_access =~ s/\".*//;
			$ctx->setreply(550,'5.7.1',$host_access);
			syslog('notice','%s/%s 550 <%s> <%s> access_db "%s"',
			       $ip_addr, $hostname, $clean_sender, $recip,
			       $host_access);
			return SMFIS_REJECT;
		    }
		    elsif (behavior('access_db',$end_user) eq 'tag')
		    {
			my $tag_name = 'access_db';
			my $tagset_href = $ctx->getpriv();
			my %tags = ();
			if (defined $$tagset_href{'tags'})
			{
			    %tags = %{$$tagset_href{'tags'}};
			}
			$tags{$tag_name} = 1;
			$$tagset_href{'tags'} = \%tags;
			$ctx->setpriv($tagset_href);
		    }
		}
	    }
	}
    }

    # Client_resolve check

    print "Checking client_resolve status of $$href{'client_resolve'}\n" if DEBUG;
    if ($$href{'client_resolve'} ne 'OK')
    {
	my $resolve_status = lc $$href{'client_resolve'};
	my $this_condition = behavior("client_resolve_$resolve_status",
				      $end_user);
	print "Our expected behavior for '$resolve_status' is '$this_condition'\n" if DEBUG;
	if ($this_condition eq 'reject')
	{
	    print "We're supposed to reject these...\n" if DEBUG;
	    if ($resolve_status eq 'temp')
	    {
                my $msg = "Connection from $ip_addr rejected due to temporary failure looking up IP address.";
                $ctx->setreply('450','4.7.1',$msg);
		        if ($SPAMLOG_ALL_TMPFAILS)
		        {
		            syslog('notice','%s/%s 450 <%s> <%s> resolve_temp "%s"',
		                   $ip_addr, $hostname, $clean_sender, $recip, $msg);
		        }
		        return SMFIS_TEMPFAIL;
	    }
	    elsif ($resolve_status eq 'forged')
	    {
		my $msg = "Connection from $ip_addr rejected due to possible forward/reverse lookup mismatch.";
		if (!$CLIENT_RESOLV_ALWAYS_TMPFAIL)
		{
		    $ctx->setreply(550,'5.7.1',$msg);
		    syslog('notice','%s/%s 550 <%s> <%s> resolve_forged "%s"',
			   $ip_addr, $hostname, $clean_sender, $recip, $msg);
		    return SMFIS_REJECT;
		}
		else
		{
		    $ctx->setreply(450,'4.7.1',$msg);
		    if ($SPAMLOG_ALL_TMPFAILS)
		    {
			syslog('notice','%s/%s 450 <%s> <%s> resolve_forged "%s"',
			       $ip_addr, $hostname, $clean_sender, $recip, $msg);
		    }
		    return SMFIS_TEMPFAIL;
		}
	    }
	    elsif ($resolve_status eq 'fail')
	    {
		my $msg = "Connection from $$href{'ip_addr'} rejected due to missing reverse DNS.";
		if (!$CLIENT_RESOLV_ALWAYS_TMPFAIL)
		{
		    $ctx->setreply(550,'5.7.1',$msg);
		    syslog('notice','%s/%s 550 <%s> <%s> resolve_missing "%s"', 
			   $ip_addr, $hostname, $clean_sender, $recip, $msg);
		    return SMFIS_REJECT;
		}
		else
		{
		    $ctx->setreply(450,'4.7.1',$msg);
		    if ($SPAMLOG_ALL_TMPFAILS)
		    {
			    syslog('notice','%s/%s 450 <%s> <%s> resolve_missing "%s"', 
			       $ip_addr, $hostname, $clean_sender, $recip, $msg);
		    }
		    return SMFIS_TEMPFAIL;
		}
	    }
	}
	elsif ($this_condition eq 'tag')
	{
	    my $tag_name = $this_condition;
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	}
    }

# Connect IP Blacklists

    if (behavior('connect_ip_dnsbl',$end_user) ne 'ignore')
    {
	my $reversed_ip = join '.', (reverse(split /\./, $ip_addr));
	foreach my $zone (keys %{$rcpt_config{'connect_ip_zones'}})
	{
	    my $resolve = $reversed_ip . '.' . $zone;
	    my ($name,$aliases,$addrtype,$length,@addrs) = 
		gethostbyname($resolve);
	    if ($#addrs > -1)
	    {
		# Bad site! Bad!
		print "$ip_addr has an A record in $zone.\n" if DEBUG;
		my $add_zone = $zone;
		$add_zone =~ s/\..*//;
		
		print "We'll rejecting this.\n" if DEBUG;
		# Rejection Code
		my $reason = "Message rejected because the IP address ($ip_addr) $rcpt_config{'connect_ip_zones'}{$zone}";

		if ( behavior('connect_ip_dnsbl',$end_user) eq 'reject')
		{
		    $ctx->setreply(550,"5.7.1",$reason);
		    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		    syslog('notice','%s/%s 550 <%s> <%s> connect_ip_dnsbl "%s"',
			   $ip_addr,$hostname,$clean_sender,$recip,$reason);
		    return SMFIS_REJECT;
		}   
		elsif ( behavior('connect_ip_dnsbl',$end_user) eq 'tag')
		{
		    my $tag_name = "connect_ip_dnsbl/$add_zone";
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}
    }

# Connect Hostname Blacklists

    if ( (behavior('connect_hostname_dnsbl',$end_user) ne 'ignore') and
	 ($hostname =~ /\w/) )
    {
	my $hostname_to_resolve = $hostname;
	foreach my $zone (keys %{$rcpt_config{'connect_host_zones'}})
	{
	    my $resolve = $hostname_to_resolve . '.' . $zone;
	    my ($name,$aliases,$addrtype,$length,@addrs) = 
		gethostbyname($resolve);
	    if ($#addrs > -1)
	    {
		# Bad site! Bad!
		print "$hostname has an A record in $zone.\n" if DEBUG;
		my $add_zone = $zone;
		$add_zone =~ s/\..*//;
		
		print "We'll reject this.\n" if DEBUG;
		# Rejection Code
		my $reason = "Message rejected because the connecting host ($hostname) $rcpt_config{'connect_host_zones'}{$zone}";

		if ( behavior('connect_hostname_dnsbl',$end_user) eq 'reject')
		{
		    $ctx->setreply(550,"5.7.1",$reason);
		    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		    syslog('notice','%s/%s 550 <%s> <%s> connect_hostname_dnsbl "%s"',
			   $ip_addr, $hostname, $clean_sender, $recip,$reason);
		    return SMFIS_REJECT;
		}   
		elsif ( behavior('connect_hostname_dnsbl',$end_user) eq 'tag')
		{
		    my $tag_name = "connect_hostname_dnsbl/$add_zone";
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}
    }

# Sender Bad NS set check

    my $rhs = $clean_sender;
    $rhs =~ s/.*\@//;

    my ($is_bad,$bad_ns) = is_bad_ns($rhs,@{$rcpt_config{'sender_bad_ns'}});

    if ($is_bad)
    {
	my $condition = 'sender_bad_ns';
	my $reason = "The sender address is in a domain ($rhs) which has a nameserver ($bad_ns) we don't like";

	print "We got to the rejection code for $condition" if DEBUG;
	if ( behavior($condition,$end_user) eq 'reject' )
	{
	    $ctx->setreply(550,"5.7.1",$reason);
	    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
	    syslog('notice','%s/%s 550 <%s> <%s> sender_bad_ns "%s"',
		   $ip_addr, $hostname, $clean_sender, $recip, $reason);

	    return SMFIS_REJECT;
	}
	elsif ( behavior($condition,$end_user) eq 'tag')
	{
	    my $tag_name = $condition;
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	}
    }

    ### Check Sender RHS criteria against bl's.

    print "Checking ${clean_sender}'s RHS of '$rhs'\n" if DEBUG;
    
    foreach my $zone (keys %{$rcpt_config{'rhs_zones'}})
    {
	my $resolve = "$rhs.$zone";
	my ($name,$aliases,$addrtype,$length,@addrs) =  gethostbyname($resolve);
	if ($#addrs > -1)
	{
	    print "$rhs has an A record in $zone.\n" if DEBUG;
	    my $add_zone = $zone;
	    $add_zone =~ s/\..*//;
	    
	    print "We'll reject this.\n" if DEBUG;
	    # Rejection Code
	    my $reason = "Message rejected because the sender's domain ($rhs) $rcpt_config{'rhs_zones'}{$zone}.";

	    if ( behavior('sender_rhs',$end_user) eq 'reject')
	    {
		$ctx->setreply(550,"5.7.1",$reason);
		print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		syslog('notice','%s/%s 550 <%s> <%s> sender_rhs "%s"',
		       $ip_addr, $hostname, $clean_sender, $recip, $reason);
		return SMFIS_REJECT;
	    }   
	    elsif ( behavior('sender_rhs',$end_user) eq 'tag')
	    {
		my $add_zone = $$href{'sender_rhs_zone'};
		
		my $tag_name = "sender_rhs/$add_zone";
		my $tagset_href = $ctx->getpriv();
		my %tags = ();
		if (defined $$tagset_href{'tags'})
		{
		    %tags = %{$$tagset_href{'tags'}};
		}
		$tags{$tag_name} = 1;
		$$tagset_href{'tags'} = \%tags;
		$ctx->setpriv($tagset_href);
	    }
	}
    }

## Connect bad NS check 

    ($is_bad,$bad_ns) = is_bad_ns($hostname,@{$rcpt_config{'connect_bad_ns'}});

    if ( 
	 ( $is_bad ) 
	 and
	 ( !whitelisted($hostname,'file'=>access_file($end_user,'whitelist')) )
	 )
	 
    {
	my $condition = 'connect_bad_ns';
	my $reason = "Your host $hostname has an NS ($bad_ns) we don't like";

	print "We got to the rejection code for $condition ($reason)\n" if DEBUG;
	if ( behavior($condition,$end_user) eq 'reject' )
	{
	    $ctx->setreply("550","5.7.1",$reason);
	    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
	    syslog('notice','%s/%s 550 <%s> <%s> connect_bad_ns "%s"',
		   $ip_addr, $hostname, $clean_sender, $recip, $reason);
	    return SMFIS_REJECT;
	}
	elsif ( behavior($condition,$end_user) eq 'tag')
	{
	    my $tag_name = $condition;
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	}
    }


# RHS Bad MX post-check handling

    if (defined $$href{'mx_forbidden'})
    {
	my @badmx_rc = @{$$href{'mx_forbidden'}};
	print "We got to the rejection code for mx_forbidden.\n" if DEBUG;
	my $reason = $badmx_rc[1];
	my $response_code = 551;
	if ($badmx_rc[0] == 2)
	{
	    $response_code = 451;
	}

	if ( behavior('mx_forbidden',$end_user) eq 'reject' )
	{
	    $ctx->setreply($response_code,'5.7.1',$reason);
	    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
	    syslog('notice','%s/%s %s <%s> <%s> mx_forbidden "%s"',
		   $ip_addr, $hostname, $response_code,
		   $clean_sender, $recip, $reason);
	    return SMFIS_REJECT if $response_code == 551;
	    return SMFIS_TEMPFAIL if $response_code == 451;
	}
	elsif ( behavior('mx_forbidden',$end_user) eq 'tag')
	{
	    my $tag_name = 'mx_forbidden';
	    $tag_name .= '_tmp' if $response_code == 451;
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	}
    }


# Freemail post-check handling

    if (defined $$href{'freemail'})
    {
	print "We got to the actual rejection code for freemail.\n" if DEBUG;
	my $reason = $$href{'freemail'};
	my $acceptable = 0;
	if (defined $users_with_forwards{$recip})
	{
	    print "This user has a forward-set. Let's check it out.\n" if DEBUG;
	    my $connected_host = $$href{'host'};
	    my @extra_hosts = @{$users_with_forwards{$recip}};
	    foreach my $test_host (@extra_hosts)
	    {
		$acceptable = 1 if $connected_host =~ /$test_host$/i;
	    }
	}
	if (! $acceptable)
	{
	    if (behavior('freemail',$end_user) eq 'reject')
	    {
		$ctx->setreply('551','5.7.1',$reason);
		print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		syslog('notice','%s/%s 551 <%s> <%s> freemail "%s"',
		       $ip_addr, $hostname, $clean_sender, $recip, $reason);
		return SMFIS_REJECT;
	    }
	    elsif ( behavior('freemail',$end_user) eq 'tag')
	    {
		my $tag_name = "freemail";
		my $tagset_href = $ctx->getpriv();
		my %tags = ();
		if (defined $$tagset_href{'tags'})
		{
		    %tags = %{$$tagset_href{'tags'}};
		}
		$tags{$tag_name} = 1;
		$$tagset_href{'tags'} = \%tags;
		$ctx->setpriv($tagset_href);
	    }
	}
	else
	{
	    print "Actually, this was coming from the ${recip}'s forwarder. It's ok. :)\n" if DEBUG;
	    my $tag_name = "freemail";
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    delete $tags{$tag_name} if defined $tags{$tag_name};
	    if (%tags)
	    {
		$$tagset_href{'tags'} = \%tags;
	    }
	    else
	    {
		delete $$tagset_href{'tags'};
	    }
	    $ctx->setpriv($tagset_href);
	}
    }

    if (defined $$href{'helo_not_resolve'})
    {
	if (behavior('helo_not_resolve',$end_user) eq 'reject')
	{
	    my $reason = $$href{'helo_not_resolve'};
	    $ctx->setreply('501','5.7.1',$reason);
	    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
	    syslog('notice','%s/%s 501 <%s> <%s> helo_not_resolve "%s"',
		   $ip_addr, $hostname, $clean_sender, $recip, $reason);
	    return SMFIS_REJECT;
	}
	elsif (behavior('helo_not_resolve',$end_user) eq 'tag')
	{
	    my $tag_name = "helo_not_resolve";
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	}
    }


    if (defined $$href{'helo_argument'})
    {
	if (behavior('helo_argument',$end_user) eq 'reject')
	{
	    my $reason = $$href{'helo_argument'};
	    $ctx->setreply("501","5.7.1",$reason);
	    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
	    syslog('notice','%s/%s 501 <%s> <%s> helo_argument "%s"',
		   $ip_addr, $hostname, $clean_sender, $recip, $reason);
	    return SMFIS_REJECT;
	}
	elsif (behavior('helo_argument',$end_user) eq 'tag')
	{
	    my $tag_name = "helo_argument";
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	}
    }

    if (defined $$href{'known_bad_helo'})
    {
	if (behavior('known_bad_helo',$end_user) eq 'reject')
	{
	    my $reason = $$href{'known_bad_helo'};
	    $ctx->setreply("501","5.7.1",$reason);
	    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
	    syslog('notice','%s/%s 501 <%s> <%s> helo_argument "%s"',
		   $ip_addr, $hostname, $clean_sender, $recip, $reason);
	    return SMFIS_REJECT;
	}
	elsif (behavior('known_bad_helo',$end_user) eq 'tag')
	{
	    my $tag_name = 'known_bad_helo';
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	}
    }

        
    
    print "+ RETURNING FROM RCPT\n" if DEBUG;
    return SMFIS_CONTINUE;
}

sub header_callback : locked
{
    my $ctx = shift;
    my $headerf = shift;
    my $headerv = shift;
    $headerv =~ s/^\s*//;
    print "+ header_callback\n" if DEBUG;
#    print "-- '$headerf' '$headerv' --\n" if DEBUG;

    my $tmp_href = $ctx->getpriv();
    $ctx->setpriv($tmp_href);
    my $recip = $$tmp_href{'already_got_a_rcpt'};
    my $clean_sender = $$tmp_href{'clean_sender'};
    my $ip_addr = $$tmp_href{'ip_addr'};
    my $hostname = $$tmp_href{'host'};
    my $end_user = $$tmp_href{'end_user'};

    my %rcpt_config = user_config($end_user,'*');
    print "Checking headers in a message for '$end_user'\n" if DEBUG;
    if (behavior('bad_headers',$end_user) ne 'ignore')
    {
	my @bad_headers = user_config($recip,'bad_headers');
	foreach my $bh (@bad_headers)
	{
	    my $compare_to = "$headerf: $headerv";
	    if ($bh =~ /^Field\:/)
	    {
		$bh =~ s/^Field\://;
		$compare_to = $headerf;
	    }
	    elsif ($bh =~ /^Value\:/)
	    {
		$bh =~ s/^Value\://;
		$compare_to = $headerv;
	    }
	    if ($compare_to =~ /($bh)/i)
	    {
		my $match = $1;
		if (behavior('bad_headers',$recip) eq 'reject')
		{
		    print "We'll reject this.\n" if DEBUG;
		    # Rejection Code
		    my $reason = "Message contains a header ($match) that is a spam-sign";
		    $ctx->setreply("550","5.7.1",$reason);
		    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		    syslog('notice','%s/%s 550 <%s> <%s> bad_headers "%s"',
			   $ip_addr, $hostname, $clean_sender, $recip, 
			   $reason);
		    return SMFIS_REJECT;
		}
		else
		{
		    my $tag_name = 'bad_headers';
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}
    }

    if ((lc $headerf ne 'from') and 
	(lc $headerf ne 'reply-to') and
	(lc $headerf ne 'subject') and
	(lc $headerf ne 'message-id')
	)
    {
	print "+ RETURNING FROM HEADER\n" if DEBUG;
	return SMFIS_CONTINUE;
    }
    
    print "Header we care about '$headerf', '$headerv'\n" if DEBUG;
    
    if ( (lc $headerf eq 'message-id') and
	 ( behavior('bad_message_id',$end_user) eq 'ignore')
	 )
    {
	# We found a message-id, but we don't really care if it's formatted
	# properly.

	my $href = $ctx->getpriv();
	$$href{'found_message_id'} = 1;
	$ctx->setpriv($href);

	print "+ RETURNING FROM HEADER\n" if DEBUG;
	return SMFIS_CONTINUE;
    }

    if ( ( lc $headerf eq 'subject') and
	 ( behavior('known_bad_subjects',$end_user) eq 'ignore')
	 )
    {
	# We got a subject, but we don't care to actually check it.
	print "+ $end_user has a preference to ignore bad subjects.\n" if DEBUG;
	print "+ RETURNING FROM HEADER\n" if DEBUG;
	return SMFIS_CONTINUE;
    }

# Bad Message-ID Check

    if (
	(lc $headerf eq 'message-id') and
	(behavior('bad_message_id',$end_user) ne 'ignore')
	)
    {
	my $href = $ctx->getpriv();
	$$href{'found_message_id'} = 1;
	$ctx->setpriv($href);

	if ($headerv !~ /$valid_message_id_regex/)
	{
	    print "Message-ID '$headerv' did not match the regex.\n" if DEBUG;

	    if (behavior('bad_message_id',$recip) eq 'reject')
	    {
		print "We'll reject this.\n" if DEBUG;
		# Rejection Code
		my $reason = "Message rejected because it contains a Message-ID header which does not appear RFC compliant";
		$ctx->setreply("550","5.7.1",$reason);
		print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		syslog('notice','%s/%s 550 <%s> <%s> bad_message_id "%s"',
		       $ip_addr, $hostname, $clean_sender, $recip, $reason);
		return SMFIS_REJECT;
	    }
	    else
	    {
		my $tag_name = 'bad_message_id';
		my $tagset_href = $ctx->getpriv();
		my %tags = ();
		if (defined $$tagset_href{'tags'})
		{
		    %tags = %{$$tagset_href{'tags'}};
		}
		$tags{$tag_name} = 1;
		$$tagset_href{'tags'} = \%tags;
		$ctx->setpriv($tagset_href);
	    }
	}
    }


# Bad NS Checks
	
    if (
	(
	 (lc $headerf eq 'from') and
	 (behavior('from_bad_ns',$end_user) ne 'ignore')
	 )
	or
	(
	 (lc $headerf eq 'reply-to') and
	 (behavior('reply_to_bad_ns',$end_user) ne 'ignore')
	 )
	)
    {
	my $address = $headerv;
	my $domain;
	{
	    lock $regex_lock;
	    ($domain) = $address =~ m/\@([A-Za-z0-9\.\-]+)/;
	}
	if (defined $domain)
	{
	    my ($is_bad,$bad_ns) = is_bad_ns($domain,
					     user_config($recip,'header_bad_ns')
					     );
	    if ($is_bad)
	    {
		my $which_behavior = lc "${headerf}_bad_ns";
		$which_behavior =~ s/\-/\_/g;
		if (behavior($which_behavior,$end_user) eq 'reject')
		{
		    print "We'll reject this.\n" if DEBUG;
		    # Rejection Code
		    my $reason = "Message rejected because it contains a $headerf value with a domain ($domain) who has an NS ($bad_ns) we don't like.";
		    $ctx->setreply("550","5.7.1",$reason);
		    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		    syslog('notice','%s/%s 550 <%s> <%s> $which_behavior "%s"',
			   $ip_addr, $hostname, $clean_sender, $recip,$reason);
		    return SMFIS_REJECT;
		}
		else
		{
		    my $tag_name = $which_behavior;
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}
    }

    if ( 
	 ( 
	   (lc $headerf eq 'from') and 
	   (behavior('from_headers',$end_user) ne 'ignore')
	   )
	 or 
	 ( 
	   (lc $headerf eq 'reply-to') and
	   (behavior('reply_to_headers',$end_user) ne 'ignore')
	   )
	 )
    {

        # RHSBL Checks

	my $address = $headerv;
	my $domain;
	{
	    lock $regex_lock;
	    ($domain) = $address =~ m/\@([A-Za-z0-9\.\-]+)/;
	}
	
	print "+ RETURNING FROM HEADER\n" if (DEBUG) and (! defined $domain);
	return SMFIS_CONTINUE if ! defined $domain;
	return SMFIS_CONTINUE 
	    if whitelisted($domain,
			   'file'=>access_file($end_user,'whitelist')
			   ); 

	foreach my $zone (keys %{$rcpt_config{'header_dom_zones'}})
	{
	    my $resolve = "$domain.$zone";
	    my ($name,$aliases,$addrtype,$length,@addrs) = gethostbyname($resolve);
	    if ($#addrs > -1)
	    {
		print "$domain has an A record in $zone.\n" if DEBUG;
		my $add_zone = $zone;
		$add_zone =~ s/\..*//;
		
		my $which_behavior = "${headerf}_headers";
		$which_behavior =~ s/\-/\_/g;
		if (behavior($which_behavior,$end_user) eq 'reject')
		{
		    print "We'll reject this.\n" if DEBUG;
		    # Rejection Code
		    my $reason = "Message rejected because it contains a $headerf header that $rcpt_config{'header_dom_zones'}{$zone}";
		    
		    $ctx->setreply("550","5.7.1",$reason);
		    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		    syslog('notice','%s/%s 550 <%s> <%s> $which_behavior "%s"',
			   $ip_addr, $hostname, $clean_sender, $recip,$reason);
		    return SMFIS_REJECT;
		}
		else
		{
		    my $tag_name = "$headerf/$add_zone";
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}
    }
    elsif (lc $headerf eq 'subject')
    {
	print "Checking subject against bad subjects...\n" if DEBUG;
	my $value = $headerv;
	foreach my $check_regex (@{$rcpt_config{'known_bad_subjects'}})
	{
	    print "Checking $check_regex...\n" if DEBUG;
	    if ($value =~ /$check_regex/)
	    {
		print "Subject regex checking found a match!\n" if DEBUG;
		if (behavior('known_bad_subjects',$end_user) eq 'reject')
		{
		    print "We'll reject this.\n" if DEBUG;
		    # Rejection Code
		    my $reason = "Message rejected because it contains a subject ('$value') that appears to be spam in nature.";
		    $ctx->setreply("550","5.7.1",$reason);
		    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		    syslog('notice','%s/%s 550 <%s> <%s> known_bad_subjects "%s"',
			   $ip_addr, $hostname, $clean_sender, $recip,
			   $reason);
		    return SMFIS_REJECT;
		}
		else
		{
		    my $tag_name = "$headerf/known_bad_subjects";
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}
    }

    print "+ RETURNING FROM HEADER\n" if DEBUG;
    return SMFIS_CONTINUE;
}



sub body_callback : locked
{
    print "+ body callback\n" if DEBUG;
    my $ctx = shift;
    my $body_segment = shift;
    my $length = shift;

    my $host_part;
    my $url;

    my $tmp_href = $ctx->getpriv();
    $ctx->setpriv($tmp_href);
    my $recip = $$tmp_href{'already_got_a_rcpt'};
    my $clean_sender = $$tmp_href{'clean_sender'};
    my $ip_addr = $$tmp_href{'ip_addr'};
    my $hostname = $$tmp_href{'host'};
    my $end_user = $$tmp_href{'end_user'};
    my %rcpt_config = user_config($recip,'*');

    if ( behavior('body_bad_regex',$end_user) ne 'ignore')
    {
	my $body_copy = $body_segment;

#	print "Checking '$body_copy' for regexes...\n" if DEBUG;

	foreach my $regex (@{$rcpt_config{'body_bad_regexes'}})
	{
	    if ($body_copy =~ /($regex)/is)
	    {
		my $matching_bit = $1;
		print "This body matched the regex '$regex'";
		if (behavior('body_bad_regex',$end_user) eq 'reject')
		{
		    my $reason = "Message rejected because it contains a string ($matching_bit) commonly found in spam or worm";
		    $ctx->setreply("550","5.7.1",$reason);
		    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		    syslog('notice','%s/%s 550 <%s> <%s> body_bad_regex "%s"',
			   $ip_addr, $hostname, $clean_sender, $recip,$reason);
		    return SMFIS_REJECT;
		}
		elsif (behavior('body_bad_regex',$end_user) eq 'tag')
		{
		    my $tag_name = 'body_bad_regex';
		    my $tagset_href = $ctx->getpriv();
		    my %tags = ();
		    if (defined $$tagset_href{'tags'})
		    {
			%tags = %{$$tagset_href{'tags'}};
		    }
		    $tags{$tag_name} = 1;
		    $$tagset_href{'tags'} = \%tags;
		    $ctx->setpriv($tagset_href);
		}
	    }
	}
    }

    my $working_copy = $body_segment;
    while ($working_copy =~ m{(https?\://([^/\s]+)/\S*)})
    {
	$url = $1;
	$host_part = $2;
	$working_copy = $POSTMATCH;
	$host_part =~ s/\:\d+$//;

	my $ip_to_check; my $hostname_to_check;
	
	print "URL: $url\n" if DEBUG;
	print "Host Part: $host_part\n" if DEBUG;
	my $obfuscated = 0;
	if ($host_part =~ /^\d+$/)
	{
	    $obfuscated = 1;
	    # 32 Bit packed address
	    my $ip = join ".", unpack "C*", pack "N", $host_part;
	    my $iaddr = inet_aton($ip);
	    my $tmp = gethostbyaddr($iaddr, AF_INET);
	    if ($tmp !~ /^\[/)
	    {
		$hostname_to_check = $tmp;
	    }
	    $ip_to_check = $ip;
	}
	elsif ($host_part =~ /^(?:\d+\.){3}\d+$/)
	{
	    # IP Address in the host part
	    
	    my $iaddr = inet_aton($host_part);
	    my $tmp = gethostbyaddr($iaddr, AF_INET);
	    if ($tmp !~ /^\[/)
	    {
		$hostname_to_check = $tmp;
	    }
	    $ip_to_check = $host_part;
	}
	else
	{
	    # Something else, must be a hostname
	    $hostname_to_check = $host_part;
	    my $iaddr = gethostbyname($hostname_to_check);
	    if (defined $iaddr)
	    {
		$ip_to_check = inet_ntoa($iaddr);
	    }
	}
	
	if (DEBUG)
	{
#		print "IP: $ip_to_check\n";
#		print "Hostname: $hostname_to_check\n";
	}
	
	if ( ($obfuscated) and
	     (behavior('obfuscated_urls',$end_user) ne 'ignore')
	     )
	{
	    print "Obfuscated!\n" if DEBUG;
	    if (behavior('obfuscated_urls',$end_user) eq 'reject')
	    {
		my $reason = "Message rejected because the body contains an URL that appears intentionally obfuscated.";
		$ctx->setreply("550","5.7.1",$reason);
		print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		syslog('notice','%s/%s 550 <%s> <%s> obfuscated_urls "%s"',
		       $ip_addr, $hostname, $clean_sender, $recip, $reason);
		return SMFIS_REJECT;
	    }
	    elsif (behavior('obfuscated_urls',$end_user) eq 'tag')
	    {
		my $tag_name = 'obfuscated_url';
		my $tagset_href = $ctx->getpriv();
		my %tags = ();
		if (defined $$tagset_href{'tags'})
		{
		    %tags = %{$$tagset_href{'tags'}};
		}
		$tags{$tag_name} = 1;
		$$tagset_href{'tags'} = \%tags;
		$ctx->setpriv($tagset_href);
	    }
	}
	
	if (defined $hostname_to_check)
	{
	    print "Checking '$hostname_to_check'\n" if DEBUG;
	    my $hostname_wl_access = lookup($hostname_to_check,
					    'type' => 'hostname',
					    'file' => access_file($end_user,
								  'whitelist')
					    );
		
	    print "... ${end_user}'s access file says '$hostname_wl_access'\n" if DEBUG;

	    if ( 
		 (behavior('url_bad_ns',$end_user) ne 'ignore' ) 
		 and
		 ( (! defined $hostname_wl_access) or 
		   ($hostname_wl_access ne 'OK')
		   )
		 )
	    {
		my ($is_bad,$bad_ns,$bad_domain) = 
		    is_bad_ns($hostname_to_check,@{$rcpt_config{'url_bad_ns'}});
		if ( ($is_bad) 
		     and 
		     (!whitelisted($bad_domain,
				   'file'=>access_file($end_user,'whitelist')))
		     and
		     (!whitelisted($bad_ns,
				   'file'=>access_file($end_user,'whitelist')))
		     )
		{
		    print "URL has bad NS!\n" if DEBUG;
		    if (behavior('url_bad_ns',$end_user) eq 'reject')
		    {
			my $reason = "Message rejected because the body contains a domain within a URL ($bad_domain) which has an NS ($bad_ns) that we don't like";
			$ctx->setreply("550","5.7.1",$reason);
			print "Rejecting mail for reason: '$reason'\n" if DEBUG;
			syslog('notice','%s/%s 550 <%s> <%s> url_bad_ns "%s"',
			       $ip_addr, $hostname, $clean_sender, $recip, 
			       $reason);
			return SMFIS_REJECT;
		    }
		    elsif (behavior('url_bad_ns',$end_user) eq 'tag')
		    {
			my $tag_name = 'url_bad_ns';
			my $tagset_href = $ctx->getpriv();
			my %tags = ();
			if (defined $$tagset_href{'tags'})
			{
			    %tags = %{$$tagset_href{'tags'}};
			}
			$tags{$tag_name} = 1;
			$$tagset_href{'tags'} = \%tags;
			$ctx->setpriv($tagset_href);
		    }
		}

	    }

	    if ( 
		 ( behavior('spam_urls',$end_user) ne 'ignore')
		 and
		 ( ( ! defined $hostname_wl_access) or
		   ($hostname_wl_access ne 'OK')
		   )
		 )
	    {
		foreach my $zone (keys %{$rcpt_config{'url_dom_zones'}})
		{
		    my $resolve = "$hostname_to_check.$zone";
		    my ($name,$aliases,$addrtype,$length,@addrs) = 
			gethostbyname($resolve);
		    if ($#addrs > -1)
		    {
			print "$hostname_to_check has a domain-based A record in $zone.\n" if DEBUG;
			if (behavior('spam_urls',$end_user) eq 'reject')
			{
			    print "We'll reject this.\n" if DEBUG;
			    # Rejection Code
			    my $reason = "Message rejected because the body contains a domain in an URL ($hostname_to_check) that $rcpt_config{'url_dom_zones'}{$zone}";
			    $ctx->setreply("550","5.7.1",$reason);
			    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
			    syslog('notice','%s/%s 550 <%s> <%s> spam_urls "%s"',
				   $ip_addr, $hostname, $clean_sender, $recip,
				   $reason);
			    return SMFIS_REJECT;
			}
			else
			{
			    my $add_zone = $zone;
			    $add_zone =~ s/\..*//;
			    
			    my $tag_name = "url/$add_zone";
			    my $tagset_href = $ctx->getpriv();
			    my %tags = ();
			    if (defined $$tagset_href{'tags'})
			    {
				%tags = %{$$tagset_href{'tags'}};
			    }
			    $tags{$tag_name} = 1;
			    $$tagset_href{'tags'} = \%tags;
			    $ctx->setpriv($tagset_href);
			}
		    }
		}
	    }
	}
	elsif ( ( behavior('unresolvable_urls',$end_user) ne 'ignore' ) and
		( ! defined $hostname_to_check) )
	{
	    if ( behavior('unresolvable_urls',$end_user) eq 'reject')
	    {
		my $reason = "Message rejected because the body contains an URL that has an unresolvable host part.";
		$ctx->setreply("550","5.7.1",$reason);
		print "Rejecting mail for reason: '$reason'\n" if DEBUG;
		syslog('notice','%s/%s 550 <%s> <%s> unresolvable_urls "%s"',
		       $ip_addr, $hostname, $clean_sender, $recip, $reason);
		return SMFIS_REJECT;
	    }
	    else
	    {
		my $tag_name = "unresolvable_url";
		my $tagset_href = $ctx->getpriv();
		my %tags = ();
		if (defined $$tagset_href{'tags'})
		{
		    %tags = %{$$tagset_href{'tags'}};
		}
		$tags{$tag_name} = 1;
		$$tagset_href{'tags'} = \%tags;
		$ctx->setpriv($tagset_href);
	    }
	}
	
	if (defined $ip_to_check)
	{
	    my $ip_wl_access = lookup($ip_to_check,
				      'type' => 'ip',
				      'file' => access_file($end_user,
							    'whitelist')
				      );
		
	    if ( 
		 ( behavior('spam_urls',$end_user) ne 'ignore') 
		 and
		 ( ( ! defined $ip_wl_access) or
		   ( $ip_wl_access ne 'OK') 
		   )
		 )
	    {
		foreach my $zone (keys %{$rcpt_config{'url_ip_zones'}})
		{
		    my $rev_ip_str = join '.', (reverse split /\./, $ip_to_check);
		    my $resolve = "$rev_ip_str.$zone";
		    my ($name,$aliases,$addrtype,$length,@addrs) = 
			gethostbyname($resolve);
		    if ($#addrs > -1)
		    {
			print "$hostname_to_check has an ip-based A record in $zone.\n" if DEBUG;
			if (behavior('spam_urls',$end_user) eq 'reject')
			{
			    print "We'll reject this.\n" if DEBUG;
			    my $reason = "Message rejected because the body contains a host/IP within an URL ($ip_to_check) that $rcpt_config{'url_ip_zones'}{$zone}";
			    $ctx->setreply("550","5.7.1",$reason);
			    print "Rejecting mail for reason: '$reason'\n" if DEBUG;
			    syslog('notice','%s/%s 550 <%s> <%s> spam_urls "%s"',
				   $ip_addr, $hostname, $clean_sender, $recip,
				   $reason);
			    return SMFIS_REJECT;
			}
			else
			{
			    my $add_zone = $zone;
			    $add_zone =~ s/\..*//;
			    
			    my $tag_name = "url/$add_zone";
			    my $tagset_href = $ctx->getpriv();
			    my %tags = ();
			    if (defined $$tagset_href{'tags'})
			    {
				%tags = %{$$tagset_href{'tags'}};
			    }
			    $tags{$tag_name} = 1;
			    $$tagset_href{'tags'} = \%tags;
			    $ctx->setpriv($tagset_href);
			}
		    }
		}
	    }
	}
    }
#   }

    print "+ RETURNING FROM BODY\n" if DEBUG;
    return SMFIS_CONTINUE;
}

sub eoh_callback
{
    my $ctx = shift;
    my $href = $ctx->getpriv();
    $ctx->setpriv($href);

    my $recip = $$href{'already_got_a_rcpt'};
    my $clean_sender = $$href{'clean_sender'};
    my $ip_addr = $$href{'ip_addr'};
    my $hostname = $$href{'host'};
    my $end_user = $$href{'end_user'};

    print "+ EOH\n" if DEBUG;
    if ( 
	 (! defined $$href{'found_message_id'})
	 and
	 (behavior('missing_message_id',$end_user) ne 'ignore')
	 )
    {
	if (behavior('missing_message_id',$end_user) eq 'reject')
	{
	    my $reason = "Message headers do not contain a Message-ID header";
	    $ctx->setreply('551','5.7.1',$reason);
	    syslog('notice','%s/%s 550 <%s> <%s> missing_message_id "%s"',
		   $ip_addr, $hostname, $clean_sender, $recip, $reason);
	    return SMFIS_REJECT;
	}
	else
	{
	    my $tag_name = 'missing_message_id';
	    my $tagset_href = $ctx->getpriv();
	    my %tags = ();
	    if (defined $$tagset_href{'tags'})
	    {
		%tags = %{$$tagset_href{'tags'}};
	    }
	    $tags{$tag_name} = 1;
	    $$tagset_href{'tags'} = \%tags;
	    $ctx->setpriv($tagset_href);
	    return SMFIS_CONTINUE;
	}
    }
    return SMFIS_CONTINUE;
}

sub eom_callback
{
    my $ctx = shift;

    print "+ EOM\n" if DEBUG;

    my $href = $ctx->getpriv();
    my %old_hash = %{$href};
    print "Got private data.\n" if DEBUG;
    my %new_hash = ( 'host' => $old_hash{'host'},
		     'ip_addr' => $old_hash{'ip_addr'},
		     'client_resolve' => $old_hash{'client_resolve'}
		     );

    $new_hash{'connect_bad_ns'} = $old_hash{'connect_bad_ns'} 
       if defined $old_hash{'connect_bad_ns'};
    $new_hash{'helo_argument'} = $old_hash{'helo_argument'} 
       if defined $old_hash{'helo_argument'};
    $new_hash{'known_bad_helo'} = $old_hash{'known_bad_helo'} 
       if defined $old_hash{'known_bad_helo'};

    print "Built new hash of data we need to save.\n" if DEBUG;
    $ctx->setpriv(\%new_hash);
    print "Saved hash, reset to post-HELO state.\n" if DEBUG;

    if (defined $old_hash{'tags'})
    {
	print "We have tags to add.\n" if DEBUG;
	# Tag Header
	my $tags_href = $old_hash{'tags'};
	print "Got the tags_href ($tags_href)\n" if DEBUG;
	my $tag_data = join ',' , ( keys %$tags_href );
	print "Adding X-SpamCheck: $tag_data\n" if DEBUG;
	$ctx->addheader('X-SpamCheck',$tag_data);
    }

    print "+ RETURNING FROM EOM\n" if DEBUG;
    return SMFIS_CONTINUE;
}


sub abort_callback
{
    my $ctx = shift;
    print "my_abort:\n" if DEBUG;
    my $href = $ctx->getpriv();
    my %newhash = ( 'host' => $$href{'host'}, 
		    'client_resolve' => $$href{'client_resolve'}, 
		    'ip_addr' => $$href{'ip_addr'}
		    );

    $newhash{'connect_bad_ns'} = $$href{'connect_bad_ns'} if (defined $$href{'connect_bad_ns'});
    $newhash{'helo_argument'} = $$href{'helo_argument'} if defined $$href{'helo_argument'};
    $newhash{'known_bad_helo'} = $$href{'known_bad_helo'} if defined $$href{'known_bad_helo'};
    $ctx->setpriv(\%newhash);

    print "   + private data restored to post-helo state.\n" if DEBUG;
    print "   + callback completed.\n" if DEBUG;
    return SMFIS_CONTINUE;
}

sub close_callback
{
    my $ctx = shift;
    print "my_close:\n" if DEBUG;
    
    $ctx->setpriv(undef);
    print "   + private data cleared.\n" if DEBUG;
    print "   + callback completed.\n" if DEBUG;
    
    return SMFIS_CONTINUE;
}

my %my_callbacks =
(
	'connect' =>	\&connect_callback,
	'helo' =>	\&helo_callback,
	'envfrom' =>	\&from_callback,
	'envrcpt' =>	\&rcpt_callback,
	'header' =>	\&header_callback,
	'eoh' =>	\&eoh_callback,
	'body' =>	\&body_callback,
	'eom' =>	\&eom_callback,
	'abort' =>	\&abort_callback,
	'close' =>	\&close_callback,
);

sub is_bad_ns : locked
{
    my $hostname = shift;
    my @check_bad_ns = @_;
    my $res = new Net::DNS::Resolver;
    my @host_parts = split /\./, $hostname;
    while (@host_parts)
    {
	my $query_hostname = join '.', @host_parts;
	my $query = $res->query($query_hostname, "NS");
	if ($query) 
	{
	    foreach my $rr ($query->answer) 
	    {
		next unless $rr->type eq "NS";
		my $ns_name = $rr->nsdname;
		
		if (grep /^$ns_name$/, (@check_bad_ns))
		{
		    return (1,$ns_name,$query_hostname);
		}
	    }
	}
	shift @host_parts;
    }
    return (0);
}

sub dnsblfile : locked
{
    my ($end_user) = @_;
    print "Seeing if $end_user has a dnsbl file.\n" if DEBUG;
    my $HOMEDIR = (getpwnam($end_user))[7];
    my $filename = $HOMEDIR . '/.spamcheck.dnsbls';
    print "Checking $filename\n" if DEBUG;
    if (! -r $filename)
    {
	print "Using default behavior.\n" if DEBUG;
	$filename = "$SPAMCHECK_FILES/spamcheck.dnsbls";
    }
    else
    {
	print "Using personal DNSBLs for $end_user\n" if DEBUG;
    }
    my %dnsbls = ('rhs_zones' => {},
		  'header_dom_zones' => {},
		  'connect_host_zones' => {},
		  'url_ip_zones' => {},
		  'url_dom_zones' => {},
		  'connect_ip_zones' => {}
		  );
    open FILE, "< $filename";
    my $current_dnsbl = '';
    while (<FILE>)
    {
	chomp;
	s/\#.*//gs;
	s/^\s*//gs;
	my $line = $_;
	next if ! /\w/;
	if (/\<(.*?)\>/)
	{
	    $current_dnsbl = $1;
	}
	elsif (/\<\/.*?\>/)
	{
	    $current_dnsbl = undef;
	}
	else
	{
	    my ($zone,$message) = $line =~ /(\S+)\s+\"(.*?)\"\s*$/;
	    if (defined $dnsbls{$current_dnsbl})
	    {
		my %current_hash = %{$dnsbls{$current_dnsbl}};
		$current_hash{$zone} = $message;
		$dnsbls{$current_dnsbl} = \%current_hash;
	    }
	    else
	    {
		my %current_hash = ($zone => $message);
		$dnsbls{$current_dnsbl} = \%current_hash;
	    }
	}

    }
    close FILE;
    return %dnsbls;
}

sub behavior
{
    my ($condition,$end_user) = @_;

    print "Checking condition $condition for $end_user\n" if DEBUG;
    my $HOMEDIR = (getpwnam($end_user))[7];
    my $filename = $HOMEDIR . '/.spamcheck.behavior';
    print "Checking $filename\n" if DEBUG;
    if (! -r $filename)
    {
	print "Using default behavior\n" if DEBUG;
	$filename = "$SPAMCHECK_FILES/spamcheck.behavior";
    }
    else
    {
	print "Using personal behavior for $end_user\n" if DEBUG;
    }

    my %behavior = ();
    open BEHAVIOR, "< $filename";
    while (<BEHAVIOR>)
    {
	chomp;
	s/\#.*//gs;
	next if ! /\w/;
	my ($key,$value) = split /\:\s+/;
	if (defined $value)
	{
	    $value =~ s/\s*$//;
	    $behavior{$key} = $value;
	}
    }
    close BEHAVIOR;

    if (defined $behavior{$condition})
    {
	print "Behavior preference: $behavior{$condition}\n" if DEBUG;
	return ($behavior{$condition});
    }
    else
    {
	print "No behavior preference. Defaulting to ignore.\n" if DEBUG;
	return 'ignore';
    }
}

sub user_config
{
    my ($recip,$REQUESTED_ITEM) = @_;

    # Stuff we don't let users muck about with

    my @known_bad_helos = ('megacity.org','yahoo.com');

    my @bad_ns = (
		  # Daily FYI
		  'ns1.dailyfyi.com','ns2.dailyfyi.com',
		  'ns1.fyispecials.com','ns2.fyispecials.com',
		  # Yahoo Store - 
		  'st-ns1.yahoo.com','st-ns2.yahoo.com','ns5.store.yahoo.com',
		  # Proactivewm
		  'dns1.proactivewm.com','dns2.proactivewm.com',
		  # Ralsky or whatever his name is
		  'ns1.tsumanipower.net','dns.cheliobon.com',
		  'ns2.cheliobon.com','dns3.elninojr.com',
		  'ns3.whodaonline.com','ns2.whodaonline.com',
		  '2ns.jacksorbetter.com','ns1.cb23.netservices411.com',
		  'ns3.icytundra.com','dv1.jacksorbetter.org',
		  'ns1.emailisfun.com','ns1.grindbind.com','dns1.grindbind.com'
		  );

    my @known_bad_subjects = ( 
			       '^[\(\[]?ADV\s*[\:\-\)\]]?\b',
			       '\s{7}[\(\[]?[a-zA-Z0-9]+[\)\]]?\s*$',
			       'new photos from my party!',
			       '^ADV\s+',
			       'I have visited your site...'
			       );

    my @body_bad_regexes = ( 
			     'Exchange Promotion and Privacy Protection Act',
			     'www\.removeyou\.com',
			     'www\.centralremovalservice\.com',
			     'This is a IE 6.0 patch',
			     'Today, Snowhite was turning 18',
			     'this is not spam',
			     'can\s?not be considered spam',
			     'S.\s*1618.*105th (?:US )?congress',
			     'of S\.\s*1618',
			     'to be (?:permanently )?removed',
			     'removal instructions',
			     'remove[\'\"]? in (?:the )?subject',
			     'I send you this file in order to have your advice',
			     'result of your form',
			     'result of your feedback form'
			     
			     );
    
    my @bad_headers = ( 
			'Field:x-uidl',
			'Field:BANNED',
			'Field:x-track'
			);

    # Stuff which users can decide for themselves
  
    my %dnsbl_hash = dnsblfile($recip);

    # Build up the hash of a given $recip's config

    my %user_config = (
		       'rhs_zones' => $dnsbl_hash{'rhs_zones'},
		       'header_dom_zones' => $dnsbl_hash{'header_dom_zones'},
		       'connect_bad_ns' => \@bad_ns,
		       'sender_bad_ns' => \@bad_ns,
		       'header_bad_ns' => \@bad_ns,
		       'url_bad_ns' => \@bad_ns,
		       'url_ip_zones' => $dnsbl_hash{'url_ip_zones'},
		       'url_dom_zones' => $dnsbl_hash{'url_dom_zones'},
		       'connect_host_zones' => $dnsbl_hash{'connect_host_zones'},
		       'connect_ip_zones' => $dnsbl_hash{'connect_ip_zones'},
		       'known_bad_subjects' => \@known_bad_subjects,
		       'body_bad_regexes' => \@body_bad_regexes,
		       'bad_headers' => \@bad_headers,
		       'known_bad_helos' => \@known_bad_helos
		       );
    
    if ($REQUESTED_ITEM eq '*')
    {
	return %user_config;
    }
    else
    {
	return @{$user_config{$REQUESTED_ITEM}};
    }
}

sub access_file
{
    my ($end_user,$whichlist) = @_;

    my $HOMEDIR = (getpwnam($end_user))[7];

    my $local_filename = $HOMEDIR . '/.spamcheck.' . $whichlist . '.db';
    my $master_filename = "$SPAMCHECK_FILES/spamcheck.$whichlist.db";

    print "looking up '$end_user' file default for '$whichlist'\n" if DEBUG;
    print "Checking $local_filename...\n" if DEBUG;
    if (! defined $whichlist)
    {
	return '/dev/null';
    }
    elsif (-r $local_filename)
    {
	return $local_filename;
    }
    else
    {
	return $master_filename;
    }
    print "what the hell?\n" if DEBUG;
}

sub is_valid_path : locked
{
    my $test_case = shift;
    my $ldh_str = '[A-Za-z0-9\-]*[A-Za-z0-9]';
    my $let_dig = '[A-Za-z0-9]';
    my $subdomain = qr{$let_dig(?:$ldh_str)*};
    my $domain = qr{(?:$subdomain(?:\.$subdomain)*
                     |
                     \[(?:\d+\.){3}\d+\])}x;
    my $atdomain = qr{\@$domain};
    my $adl = qr{$atdomain(?:\,$atdomain)*};
    my $atom = '[-a-zA-Z0-9!\#\$%&\'*+/=?^_`{|}~]';
    my $dotstring = qr{(?:$atom*(?:\.(?:$atom)*)*)};
    my $quotedstring = qr{\"[^\"]+\"};
    my $local_part = qr{(?:$dotstring|$quotedstring)};
    my $mailbox = qr {(?:$local_part\@$domain)};
    my $path = qr {(?: \< (?: $adl\: )? $mailbox \> ) }x;
    my $reverse_path = qr { (?: $path | \<\> ) }x;
    if ($test_case !~ /$reverse_path/)
    {
        return 0;
    }
    else
    {
        return 1;
    }
}

sub alias_scan
{
    my $address = shift;

    my %access;
    my $filename = '/etc/mail/aliases.db';
    print "Tie-ing...\n" if DEBUG;
    tie %access, 'DB_File', $filename , O_RDONLY  or die "Couldn't open $filename : $!";;
    print "Checking $address\n" if DEBUG;
    $address .= "\000";
    while (defined $access{$address})
    {
	my $new_addr = $access{$address};
	print "$address maps to '$new_addr'\n" if DEBUG;
	my $new_address = $new_addr;
	if ($new_addr =~ /\,/)
	{
	    my @addrs = (split /\,/, $new_addr);	
	    $new_address = $addrs[0] . "\000";
	}
	last if $address eq $new_address;     # handle foo: foo,bar loop
	$address = $new_address;
	if ($address =~ /\|/)
	{
	    $address = 'SHELL COMMAND';
	}
	if ($address =~ /\@/)
	{
	    $address = 'REMOTE USER';
	}
	print "Checking $address\n" if DEBUG;
    }
    untie %access;
    $address =~ s/\000//;
    return $address;
}


sub bad_mx_set : locked
{
    # rc = 0   Good
    # rc = 1   Error
    # rc = 2   Temp Failure

    my $domain = shift;
    my $res = Net::DNS::Resolver->new;
    my @mx = mx($res,$domain);
    
    my $rc = 0;
    my $rc_string = 'OK';

    my @ip_addr = ();
    
    if (! @mx)
    {
	if ( $res->errorstring eq 'NOERROR')
	{
	    print "$domain has no MX, using hostname\n" if DEBUG;
	    @mx = ($domain);
	}
	elsif ( $res->errorstring eq 'NXDOMAIN')
	{
	    print "MX lookup on $domain has an NXDOMAIN state.\n" if DEBUG;
	    return (1,"$domain does not exist");
	}
	elsif ( $res->errorstring eq 'SERVFAIL')
	{
	    $rc = 2;
	    $rc_string = "Temporarily unable to reach NS servers for $domain to do an MX lookup.";
	    @mx = ();
	    print "MX lookup on $domain has SERVFAIL state.\n" if DEBUG;
	}
	else
	{
	    $rc = 2;
	    $rc_string = $res->errorstring;
	    @mx = ();
	}
    }
    
    if ( ! $rc )
    {
	foreach my $mx_entry (@mx)
	{
	    my $pref;
	    if (ref $mx_entry)
	    {
		$pref = $mx_entry->exchange;
	    }
	    else
	    {
		$pref = $mx_entry;
	    }
	    print "Checking MX Entry: $pref\n" if DEBUG;

	    if ($pref =~ /^(\d+\.){3}\d+$/)
	    {
		return(1,"$pref is not a valid value for an MX RR. Only hostnames are valid.");
	    }
	    else
	    {
		my $query = $res->search($pref);
		if ($query)
		{
		    foreach my $rr ($query->answer) 
		    {
			next unless $rr->type eq 'A';
			push @ip_addr, ($rr->address);
			print "Pushed ", $rr->address, " into \@ip_addr\n" if DEBUG;
		    }
		}
		else
		{
		    if ($res->errorstring eq 'SERVFAIL')
		    {
			$rc = 2;
			$rc_string = "MX host $pref temporarily not confirmable.";
		    }
		    elsif ( ($res->errorstring eq 'NXDOMAIN') or
			    ($res->errorstring eq 'NOERROR') )
		    {
			my $foo = $res->errorstring;
			return (1, "MX host $pref has no A RRs ($foo)");
		    }
		    else
		    {
			$rc = 2;
			$rc_string = $res->errorstring;
		    }
		}
	    }
	}
	print "Done looking for IP addresses.\n" if DEBUG;
	if (@ip_addr)
	{
	    foreach my $addr (@ip_addr)
	    {
		print "Checking IP $addr\n" if DEBUG;
		foreach my $net (@forbidden)
		{
		    print "Checking Network $net\n" if DEBUG;
		    if (Net::CIDR::cidrlookup($addr,$net))
		    {
			return (1, "$domain has an MX with an IP in reserved space ($addr is in $net)");
		    }
		}
	    }
	}

    }

    return ($rc,$rc_string);
}



BEGIN:
{
    if (scalar(@ARGV) < 2)
    {
	print "Usage: perl $0 <name_of_filter> <path_to_sendmail.cf>\n";
	exit;
    }
    
    my $conn = Sendmail::Milter::auto_getconn($ARGV[0], $ARGV[1]);
    
    open RUN, "> $PID_FILE";
    print RUN "$$\n";
    autoflush RUN 1;
    close RUN;
    
    setlogsock('unix');
    openlog ('spamcheck','cons,pid',$FACILITY);

    print "Found connection info for '$ARGV[0]': $conn\n" if DEBUG;
    
    if ($conn =~ /^local:(.+)$/)
    {
	my $unix_socket = $1;
	
	if (-e $unix_socket)
	{
	    print "Attempting to unlink UNIX socket '$conn' ... " if DEBUG;
	    
	    if (unlink($unix_socket) == 0)
	    {
		print "failed.\n" if DEBUG;
		unlink $PID_FILE;
		exit;
	    }
	    print "successful.\n" if DEBUG;
	}
    }
    
    if (not Sendmail::Milter::auto_setconn($ARGV[0], $ARGV[1]))
    {
	print "Failed to detect connection information.\n";
	unlink $PID_FILE;
	exit;
    }
    
    #
    #  The flags parameter is optional. SMFI_CURR_ACTS sets all of the
    #  current version's filtering capabilities.
    #
    #  %Sendmail::Milter::DEFAULT_CALLBACKS is provided for you in getting
    #  up to speed quickly. I highly recommend creating a callback table
    #  of your own with only the callbacks that you need.
    #
    
    if (not Sendmail::Milter::register($ARGV[0], \%my_callbacks,
				       SMFI_CURR_ACTS))
    {
	print "Failed to register callbacks for $ARGV[0].\n";
	unlink $PID_FILE;
	exit;
	}
    
    print "Starting Sendmail::Milter $Sendmail::Milter::VERSION engine.\n" if DEBUG;
    
    if (Sendmail::Milter::main())
    {
	print "Successful exit from the Sendmail::Milter engine.\n" if DEBUG;
    }
    else
    {
	print "Unsuccessful exit from the Sendmail::Milter engine.\n" if DEBUG;
    }
    unlink $PID_FILE;

}

